from typing import Union, List
from posixpath import join as urljoin
import warnings
import os

from pydantic import BaseModel
from max_ard.session import get_user_session
from max_ard.exceptions import ARDServerException, BadARDRequest, NotSubmitted

from max_ard.session import ARD_HOSTNAME, ARD_DEV_BRANCH

POST_ORDER_URL = f'https://{ARD_HOSTNAME}/order{ARD_DEV_BRANCH}'
GET_ORDER_URL = f'https://{ARD_HOSTNAME}/order/status{ARD_DEV_BRANCH}'

##############
# Validation decorators
#############


def needs_submittal(func):
    def func_wrapper(instance):
        if not instance.submitted:
            raise NotSubmitted(f"Can't access {func.__name__} until Select has been submitted")
        res = func(instance)
        return res
    return func_wrapper


#######
# Models
#######

class Acquisition(BaseModel):
    id: str
    cells: List[str] = []

    class Config:
        validate_assignment = True


class Output_Config(BaseModel):
    bucket: str 
    prefix: str 
    role_arn: str = None

    class Config:
        validate_assignment = True


class EmailNotification(BaseModel):
    type: str = 'email'
    address: str
 

    class Config:
        allow_mutation: False

class SNSNotification(BaseModel):
    type: str = 'sns'
    topic_arn: str
        
    class Config:
        allow_mutation: False


class OrderRequest(BaseModel):
    acquisitions: List[Acquisition] = None
    select_id: str = None
    dry_run: bool = False
    # filter
    intersects: str = None
    bbox:  List = None
    # destination
    output_config: Output_Config = None
    bundle_adjust: bool = False
    notifications: List[Union[EmailNotification, SNSNotification]] = []
    metadata: dict = {}

    class Config:
        extra = 'allow'
        validate_assignment = True

    def to_payload(self):
        ''' sends only parameters with values '''
        params = self.dict().copy()
        payload = {k:v for k,v in params.items() if v not in (None, [], ())}
        params = payload['output_config'].copy()
        payload['output_config'] = {k:v for k,v in params.items() if v is not None}
        return payload

class OrderResponse(BaseModel):
    id: str
    state: str
    status_message: str

    class Config:
        extra = 'allow'
        validate_assignment = True

class Order():

    def __init__(self,
        acquisitions = None,
        select_id = None,
        destination = None,
        intersects = None,
        bbox = None,
        role_arn = None,
        dry_run = False,
        bba = False,
        metadata = {},
        ):
        
    
        # set up the output config
        if destination is not None:
            destination = destination.replace('s3://', '')
            bucket, prefix = destination.split('/', 1)
            output_config = Output_Config(bucket=bucket, prefix=prefix, role_arn=role_arn)
        else:
            output_config = None
        

        # set up acquisitions
        if type(acquisitions) == str:
            acquisitions = [Acquisition(id=acquisitions)]
        elif type(acquisitions) == list or type(acquisitions) == tuple:
            if len(acquisitions) > 0:
                if type(acquisitions[0]) == str:
                    acquisitions = [Acquisition(id=v) for v in acquisitions]
        elif type(acquisitions) == dict:
            acquisitions = [Acquisition(id=k, cells=v) for k,v in acquisitions.items()]

        # if the intersects is a Shapely geom, convert it to wkt
        try:
            intersects = intersects.wkt
        except:
            pass

        self.request = OrderRequest(
            acquisitions=acquisitions,
            select_id=select_id,
            intersects=intersects,
            bbox=bbox,
            output_config=output_config,
            bundle_adjust=bba,
            dry_run=dry_run,
            metadata=metadata
        )
        self.dry_run = dry_run
        self.response = None
        self.submitted = False


    def add_email_notification(self, address):
        self.request.notifications.append(EmailNotification(address=address))

    def add_sns_notification(self, topic_arn):
        self.request.notifications.append(SNSNotification(topic_arn=topic_arn))


    @property
    @needs_submittal
    def finished(self):
        ''' The Order is finished processing and tiles are available '''

        return self.state in ['FINISHED', 'SUCCEEDED']

    @property
    @needs_submittal
    def order_id(self):
        ''' The Order ID'''

        return self.response.id

    @property
    @needs_submittal
    def state(self):
        ''' State of the order process: 'RUNNING', 'FINISHED' or 'SUCCEEDED, or 'FAILED' '''

        if self.response.state == 'RUNNING':
            response = self.get_order(self.order_id)
            self.response = OrderResponse(**response) 
        return self.response.state


    @classmethod
    def from_id(cls, order_id, session=None):
        ''' Create a Order object from an ID '''

        response = cls.get_order(order_id, session)
        instance = cls()
        instance.response = OrderResponse(**response) 
        # can recreate the request
        instance.submitted = True
        return instance

    @classmethod
    def get_order(cls, order_id, session=None):
        ''' Gets data from the API about a select '''

        if not session:
            session = get_user_session()

        url = urljoin(GET_ORDER_URL, order_id)
        r = session.get(url)
        if r.status_code == 400:
            raise BadARDRequest(r.json()['message'])
        # server error
        elif r.status_code != 200:
            raise ARDServerException(r.json()['message'])
        # success
        return r.json()


    @classmethod
    def send_order(cls, payload, session=None):
        ''' Send an order request to the API '''

        if not session:
            session = get_user_session()
        r = session.post(POST_ORDER_URL, json=payload)
        # bad format
        if r.status_code == 400:
            raise BadARDRequest(r.json()['message'])
        # server error
        elif r.status_code != 202:
            raise ARDServerException(r.json()['message'])
        # success
        return r.json()
    

    def submit(self):
        ''' Submits an order request to the API '''

        response = self.send_order(self.request.to_payload())
        if self.submitted:
            warnings.warn('The Order has already been submitted')
            return
        self.submitted = True
        if self.dry_run:
            self.response = OrderResponse(
                id='dry run',
                state='SUCCEEDED',
                status_message='Order dry-run validation successful.'
            )
        else:
            self.response = OrderResponse(
                id=response['id'],
                state='RUNNING',
                status_message='Order submitted'
            )